#include "atom_parser.hh"

#include <expat.h>

#include <iostream>

const char* const
Atom_parser::Entry::ddl("create table vtable ("
                        "title text, "
                        "link text, "
                        "id text, "
                        "updated text, "
                        "author text, "
                        "summary text"
                        ") ");

const Atom_parser::Action_map::value_type
Atom_parser::s_action_list[] = {
  Atom_parser::Action_map::value_type("author",   &Atom_parser::author_tag),
  Atom_parser::Action_map::value_type("entry",   &Atom_parser::entry_tag),
  Atom_parser::Action_map::value_type("id",      &Atom_parser::id_tag),
  Atom_parser::Action_map::value_type("link",    &Atom_parser::link_tag),
  Atom_parser::Action_map::value_type("summary", &Atom_parser::summary_tag),
  Atom_parser::Action_map::value_type("title",   &Atom_parser::title_tag),
  Atom_parser::Action_map::value_type("updated", &Atom_parser::updated_tag),
};

const Atom_parser::Action_map
Atom_parser::s_action(&s_action_list[0],
                      &s_action_list[sizeof(s_action_list)/sizeof(*s_action_list)]);

Atom_parser::Handler
Atom_parser::find_handler(const XML_Char *name) const
{
  Action_map::const_iterator it = s_action.find(name);
  return it != s_action.end() ? it->second : 0;
}


void
Atom_parser::author_tag(const XML_Char **atts,
                        Tag_type tag_type)
{
  switch (tag_type) {
  case TAG_TYPE_START:
    m_assign_to = m_is_in_entry ? &m_pending.author : &m_feed.author;
    break;

  case TAG_TYPE_END:
    m_assign_to = 0;
    break;
  }
}


void
Atom_parser::entry_tag(const XML_Char **atts,
                       Tag_type tag_type)
{
  switch (tag_type) {
  case TAG_TYPE_START:
    m_is_in_entry = true;
    m_pending.clear();
    if (!m_feed.author.empty())
      m_pending.author = m_feed.author;
    break;

  case TAG_TYPE_END:
    m_is_in_entry = false;
    m_entries.push_back(m_pending);
    break;
  }
}

void
Atom_parser::title_tag(const XML_Char **atts,
                       Tag_type tag_type)
{
  switch (tag_type) {
  case TAG_TYPE_START:
    if (m_is_in_entry)
      m_assign_to = &m_pending.title;
    break;

  case TAG_TYPE_END:
    m_assign_to = 0;
    break;
  }
}

void
Atom_parser::link_tag(const XML_Char **atts,
                      Tag_type tag_type)
{
  switch (tag_type) {
  case TAG_TYPE_START:
    if (m_is_in_entry) {
      while (*atts) {
        if (strcmp("href", *atts++) == 0) {
          m_pending.link = *atts;
          break;
        }
        atts++;
      }
    }
    break;

  case TAG_TYPE_END:
    // Do nothing
    break;
  }
}

void
Atom_parser::id_tag(const XML_Char **atts,
                    Tag_type tag_type)
{
  switch (tag_type) {
  case TAG_TYPE_START:
    if (m_is_in_entry)
      m_assign_to = &m_pending.id;
    break;

  case TAG_TYPE_END:
    m_assign_to = 0;
    break;
  }
}


void
Atom_parser::updated_tag(const XML_Char **atts,
                         Tag_type tag_type)
{
  switch (tag_type) {
  case TAG_TYPE_START:
    if (m_is_in_entry)
      m_assign_to = &m_pending.updated;
    break;

  case TAG_TYPE_END:
    m_assign_to = 0;
    break;
  }
}


void
Atom_parser::summary_tag(const XML_Char **atts,
                         Tag_type tag_type)
{
  switch (tag_type) {
  case TAG_TYPE_START:
    if (m_is_in_entry)
      m_assign_to = &m_pending.summary;
    break;

  case TAG_TYPE_END:
    m_assign_to = 0;
    break;
  }
}


void
Atom_parser::start_handler(void *data, const XML_Char *name, const XML_Char **attrs)
{
  Atom_parser *parser = reinterpret_cast<Atom_parser*>(data);
  Handler handler = parser->find_handler(name);
  if (handler != 0)
    (parser->*handler)(attrs, TAG_TYPE_START);
}


void
Atom_parser::end_handler(void *data, const XML_Char *name)
{
  Atom_parser *parser = reinterpret_cast<Atom_parser*>(data);
  Handler handler = parser->find_handler(name);
  if (handler != 0)
    (parser->*handler)(0, TAG_TYPE_END);
}

void
Atom_parser::text_handler(void *data, const XML_Char *text, int len)
{
  Atom_parser *parser = reinterpret_cast<Atom_parser*>(data);
  if (parser->m_assign_to)
    parser->m_assign_to->append(text,len);
}


Atom_parser::Atom_parser(FILE* file)
  : m_parser(0), m_file(file), m_eof(false), m_is_in_entry(false), m_assign_to(0)
{
  XML_Parser p;

  if ((p = XML_ParserCreate(NULL)) == NULL)
    return;

  XML_SetElementHandler(p, start_handler, end_handler);
  XML_SetCharacterDataHandler(p, text_handler);
  XML_SetUserData(p, this);

  m_parser = p;
}

/**
   Read at least one entry from file.
 */
int
Atom_parser::fill_from_file()
{
  while (m_entries.empty() && !m_eof) {
    char buffer[1024];
    int len = fread(buffer, 1, sizeof(buffer), m_file);
    if (ferror(m_file))
      return ATOM_ERR_FILE_READ;

    m_eof = feof(m_file);

    if (! XML_Parse(m_parser, buffer, len, m_eof)) {
      fprintf(stderr, "Parse error at line %d:\n%s\n",
	      XML_GetCurrentLineNumber(m_parser),
	      XML_ErrorString(XML_GetErrorCode(m_parser)));
      return ATOM_ERR_PARSE;
    }
  }
}

