#include "feed_vtab.h"

#include "config.h"

#include "atom_parser.hh"

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <curl/curl.h>
#include <curl/easy.h>

#include <expat.h>
#include <sqlite3ext.h>

SQLITE_EXTENSION_INIT1

/* DDL defining the structure of the vtable */
typedef struct Feed_vtab
{
  sqlite3_vtab base;
  sqlite3 *db;
  char *url;
} Feed_vtab;

struct Feed_cursor
{
  sqlite3_vtab_cursor base;
  Atom_parser *parser;
  FILE *file;
};


static int vt_destructor(sqlite3_vtab *pVtab)
{
  Feed_vtab *p = (Feed_vtab*)pVtab;
  sqlite3_free(p);

  return 0;
}

static int vt_create( sqlite3 *db, void *p_aux,
                      int argc, const char* const* argv,
                      sqlite3_vtab **pp_vt,
                      char **pzErr )
{
  int rc = SQLITE_OK;
  Feed_vtab* p_vt;

  /* Allocate the sqlite3_vtab/Feed_vtab structure itself */
  p_vt = (Feed_vtab*) sqlite3_malloc(sizeof(*p_vt));

  if(p_vt == NULL)
    return SQLITE_NOMEM;

  if (argv[3] == NULL)
    return SQLITE_ERROR;

  const char *url_arg = argv[3];
  size_t url_len = strlen(url_arg);

  if (url_arg[0] == '"' || url_arg[0] == '\'') {
    url_arg += 1;
    url_len -= 2;
  }

  p_vt->url = strndup(url_arg, url_len);
  p_vt->db = db;

  /* Declare the vtable's structure */
  rc = sqlite3_declare_vtab(db, Atom_parser::Entry::ddl);

  if(rc != SQLITE_OK)
  {
    vt_destructor((sqlite3_vtab*)p_vt);
    return SQLITE_ERROR;
  }

  /* Success. Set *pp_vt and return */
  *pp_vt = &p_vt->base;

  return SQLITE_OK;
}

static int vt_connect( sqlite3 *db, void *p_aux,
                       int argc, const char * const* argv,
                       sqlite3_vtab **pp_vt, char **pzErr )
{
  return vt_create(db, p_aux, argc, argv, pp_vt, pzErr);
}

static int vt_disconnect(sqlite3_vtab *pVtab)
{
  return vt_destructor(pVtab);
}

static int vt_destroy(sqlite3_vtab *pVtab)
{
  int rc;
  rc = vt_destructor(pVtab);
  return rc;
}

static int vt_open(sqlite3_vtab *vtab, sqlite3_vtab_cursor **pp_cursor)
{
  CURL *p_handle;
  int error;
  Feed_vtab *p_vtab = (Feed_vtab*) vtab;

  Feed_cursor *p_cur
    = (Feed_cursor*)sqlite3_malloc(sizeof(Feed_cursor));

  if (!p_cur)
    return SQLITE_NOMEM;

  p_cur->file = tmpfile();
  p_cur->parser = new Atom_parser(p_cur->file);

  if ((p_handle = curl_easy_init()) == NULL) {
    return SQLITE_NOMEM;
  }

  assert(p_handle != NULL);

  /* Setup libcurl for reading into the temporary file */
  curl_easy_setopt(p_handle, CURLOPT_URL, p_vtab->url);
  curl_easy_setopt(p_handle, CURLOPT_WRITEDATA, p_cur->file);

  /* Read the URL into file. */
  if ((error = curl_easy_perform(p_handle)))
  {
    
    return SQLITE_ERROR;
  }

  /* Clean up after reading the URL */
  curl_easy_cleanup(p_handle);

  // Rewind to start parsing at the beginning of the file
  fseek(p_cur->file, 0L, SEEK_SET);

  // Advance to first entry
  p_cur->parser->advance_entry();

  *pp_cursor = (sqlite3_vtab_cursor*)p_cur;

  return (p_cur ? SQLITE_OK : SQLITE_NOMEM);
}

static int vt_close(sqlite3_vtab_cursor *cur)
{
  Feed_cursor *p_cur = (Feed_cursor*) cur;

  fclose(p_cur->file);
  delete p_cur->parser;
  sqlite3_free(p_cur);

  return SQLITE_OK;
}

static int vt_eof(sqlite3_vtab_cursor *cur)
{
  return reinterpret_cast<Feed_cursor*>(cur)->parser->empty();
}

static int vt_next(sqlite3_vtab_cursor *cur)
{
  reinterpret_cast<Feed_cursor*>(cur)->parser->advance_entry();
  return SQLITE_OK;
}


static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int i)
{
  Feed_cursor *p_cur = (Feed_cursor*)cur;
  Atom_parser::Entry *entry = p_cur->parser->current_entry();

  switch (i) {
  case 0:                                       // title
    sqlite3_result_text(ctx, entry->title.c_str(), -1, SQLITE_STATIC);
    break;
  case 1:                                       // link
    sqlite3_result_text(ctx, entry->link.c_str(), -1, SQLITE_STATIC);
    break;
  case 2:                                       // id
    sqlite3_result_text(ctx, entry->id.c_str(), -1, SQLITE_STATIC);
    break;
  case 3:                                       // updated
    sqlite3_result_text(ctx, entry->updated.c_str(), -1, SQLITE_STATIC);
    break;
  case 4:                                       // author
    sqlite3_result_text(ctx, entry->author.c_str(), -1, SQLITE_STATIC);
    break;
  case 5:                                       // summary
    sqlite3_result_text(ctx, entry->summary.c_str(), -1, SQLITE_STATIC);
    break;
  }

  return SQLITE_OK;
}

static int vt_rowid(sqlite3_vtab_cursor *cur, sqlite_int64 *p_rowid)
{
  Feed_cursor *p_cur = (Feed_cursor*) cur;
  *p_rowid = (sqlite_int64) p_cur->parser->current_entry();
  return SQLITE_OK;
}

static int vt_filter( sqlite3_vtab_cursor *p_vtc, 
                      int idxNum, const char *idxStr,
                      int argc, sqlite3_value **argv )
{
  return SQLITE_OK;
}

static int vt_best_index(sqlite3_vtab *tab, sqlite3_index_info *pIdxInfo)
{
  return SQLITE_OK;
}

sqlite3_module feed_module = 
{
  0,              /* iVersion */
  vt_create,      /* xCreate       - create a vtable */
  vt_connect,     /* xConnect      - associate a vtable with a connection */
  vt_best_index,  /* xBestIndex    - best index */
  vt_disconnect,  /* xDisconnect   - disassociate a vtable with a connection */
  vt_destroy,     /* xDestroy      - destroy a vtable */
  vt_open,        /* xOpen         - open a cursor */
  vt_close,       /* xClose        - close a cursor */
  vt_filter,      /* xFilter       - configure scan constraints */
  vt_next,        /* xNext         - advance a cursor */
  vt_eof,         /* xEof          - inidicate end of result set*/
  vt_column,      /* xColumn       - read data */
  vt_rowid,       /* xRowid        - read data */
  NULL,           /* xUpdate       - write data */
  NULL,           /* xBegin        - begin transaction */
  NULL,           /* xSync         - sync transaction */
  NULL,           /* xCommit       - commit transaction */
  NULL,           /* xRollback     - rollback transaction */
  NULL,           /* xFindFunction - function overloading */
};

extern "C" int feed_register(sqlite3 *db)
{
  return sqlite3_create_module(db, "feedme", &feed_module, NULL);
}
